local entityModule = {}
local blocksModule = require('blocks')

entityModule.entities = {}
entityModule.functions = {}

entityModule.functions.addEntity = function(x, y, width, height, name)
    entityModule.entities[name] = {
        x=x,
        y=y,
        width=width,
        height=height
    }
end

entityModule.functions.updateEntityVelocity = function(entityName, xVel, yVel)
    if entityModule.entities[entityName] then
        local entity = entityModule.entities[entityName]

        local eX, eY = math.floor(entity.x/16)+1, math.floor(entity.y/16)+3
        local IndexOfWorld = blocksModule.world[eX]
        local block = IndexOfWorld and blocksModule.world[eX][eY] or nil

        local function updateXVel(x)
            entity.x=entity.x+xVel+x
        end
        local function updateYVel(y)
            entity.y=entity.y+yVel+y
        end

        if block and block ~= nil then  
            -- object exists
            local tileLeftOverlaps = math.floor((entity.x / blockSize)+0.5) > (block.pos.x + 0.5)
            local tileTopOverlaps = math.floor((entity.y / blockSize)+0.5) < (block.pos.y - 0.5)
            local tileBottomOverlaps = math.floor((entity.y / blockSize)+0.5) > (block.pos.y + 0.5)
            local tileRightOverlaps = math.floor((entity.x / blockSize)+0.5) < (block.pos.x - 0.5)
        
            local xOverlap = tileLeftOverlaps or tileRightOverlaps
            local yOverlap = tileTopOverlaps or tileBottomOverlaps
        
            if not xOverlap then
                updateXVel()
            end
            if not yOverlap then
                updateYVel()
            else
                if tileTopOverlaps then
                    local distance = (math.abs(entity.y)/blockSize) - (math.abs(block.pos.y) + 0.5)
                    entity.y = entity.y + (distance/blockSize)
                    return
                end
                if tileBottomOverlaps then
                    local distance = (math.abs(entity.y)/blockSize) - (math.abs(block.pos.y) - 0.5)
                    entity.y = entity.y - (distance/blockSize)
                    return
                end
            end
        else
            updateXVel(0)
            updateYVel(90*love.timer.getAverageDelta())
        end
    end
end

return entityModule