blocksModule = require('blocks')
lightingData = require('lighting') -- needs a registered lighting source in here if you want light sources.
entitiesModule = require('entities') -- dynamic objects.
physicsModule = require('physics') -- used for literally everything.

function love.load()
    blockSize = 16
    worldWidth = 2^8
    worldHeight = 2^10

    blockContainer = blocksModule.world

    lightingData.addSource(150, -50, 10, nil, {r=1,g=0,b=0})
    lightingData.addSource(170, -50, 10, nil, {r=1,g=0.5,b=0})
    lightingData.addSource(190, -50, 10, nil, {r=0,g=1,b=0})
    lightingData.addSource(210, -50, 10, nil, {r=0,g=0.5,b=1})

    cameraX, cameraY = 0, 0
    cameraRoundedX = 0
    cameraRoundedY = 0

    entitiesModule.functions.addEntity(0, 0, 16, 16, 'Player')

    screenWidth, screenHeight = love.graphics.getWidth(), love.graphics.getHeight()
    -- create a world.

    math.randomseed(os.time())
    worldSeed = math.random(1, 1000000)/1000000
    math.randomseed(worldSeed, worldSeed);
    for x = 1, worldWidth do
        blockContainer[x] = {}
        local h = 0
        for y = 1, worldHeight do
            local noise1 = love.math.noise(x/200, x*worldSeed/400) * 24
            local flatnessNoise = love.math.noise(x/4000, x*worldSeed/200) * 10
            local mountaineousNoise = love.math.noise(y/1000, x/1000*worldSeed-200)*20
    
            local noise = noise1 + flatnessNoise + mountaineousNoise
    
            -- Calculate the noise at x-1
            local xN = x-1
            local yN = y-1

            local noise1MinusOne = love.math.noise(xN/200, xN*worldSeed/400) * 24
            local flatnessNoiseMinusOne = love.math.noise(xN/4000, xN*worldSeed/200) * 10
            local mountaineousNoise = love.math.noise(yN/1000, xN/1000*worldSeed-200)*20

            local noiseNeg = noise1MinusOne + flatnessNoiseMinusOne + mountaineousNoise
            
            -- Calculate the slope
            local slope = math.abs(((noise - noiseNeg)))
            local slopeThreshold = 0.2
            if y > math.ceil(noise) then

                if h < 1 and slope < slopeThreshold then
                    local block = blocksModule.functions.createBlock(1) -- get block
                    block.state = 1
                    block.slope = slope
                    block.pos = {
                        x = x,
                        y = y
                    }
                    blockContainer[x][y] = block -- add to blocks container aka. world
                elseif h < 7 and slope < slopeThreshold then
                    local block = blocksModule.functions.createBlock(3) -- get block
                    block.state = 1
                    block.slope = slope
                    block.pos = {
                        x = x,
                        y = y
                    }
                    blockContainer[x][y] = block -- add to blocks container aka. world
                else
                    local block = blocksModule.functions.createBlock(2) -- get block
                    block.state = 1
                    block.slope = slope
                    block.pos = {
                        x = x,
                        y = y
                    }
                    blockContainer[x][y] = block -- add to blocks container aka. world
                end

                local ribbonDist = (y-320) / 2;
                local threshold = 1.0 - 0.4 * math.min(ribbonDist*ribbonDist, 1.0);
                local caveNoise = love.math.noise((x/30)*worldSeed+(worldSeed*400), (y/30)*worldSeed+(worldSeed*400), worldSeed*worldSeed)

                if caveNoise >= threshold then
                    blockContainer[x][y] = nil
                end
                --------
                local ribbonDist = (y+32) / 10;
                local threshold = 1.0 - 0.27 * math.min(ribbonDist*ribbonDist, 1.0);
                local noise = love.math.noise((x+(worldSeed*10000000))/9, (y-(worldSeed*10000000))/9, (worldSeed*10000000))

                if noise >= threshold and blockContainer[x][y] and y > 48 then
                    local block = blocksModule.functions.createBlock(6) -- get block
                    block.state = 1
                    block.slope = slope
                    block.pos = {
                        x = x,
                        y = y
                    }
                    blockContainer[x][y] = block -- add to blocks container aka. world
                end
                --------
                local ribbonDist = (y+3) / 180;
                local threshold = 1.0 - 0.27 * math.min(ribbonDist*ribbonDist, 1.0);
                local noise = love.math.noise((x+(worldSeed*10000000))/15, (y-(worldSeed*10000000))/9, (worldSeed*10000000))

                if noise >= threshold and blockContainer[x][y] and y > 150 then
                    local block = blocksModule.functions.createBlock(7) -- get block
                    block.state = 1
                    block.slope = slope
                    block.pos = {
                        x = x,
                        y = y
                    }
                    blockContainer[x][y] = block -- add to blocks container aka. world
                end
                h = h + 1
            end
        end
    end

    
    -- test block
    local block = blocksModule.functions.createBlock(4) -- get block
    block.state = 1
    blockContainer[worldWidth] = {}
    blockContainer[worldWidth][-200] = block -- add to blocks container aka. world
    --

end

function love.update(dt)
    screenWidth, screenHeight = love.graphics.getWidth(), love.graphics.getHeight()
    
    local moveX = (love.keyboard.isDown('a') and -256 or love.keyboard.isDown('d') and 256 or 0)
    local moveY = (love.keyboard.isDown('w') and -512-90 or love.keyboard.isDown('s') and 512 or 0)

    entitiesModule.functions.updateEntityVelocity('Player', moveX*dt, moveY*dt)

    cameraX = entitiesModule.entities.Player.x - (screenWidth/2)--cameraX + moveX * dt
    cameraY = -entitiesModule.entities.Player.y + (screenHeight/2)--cameraY - moveY * dt

    cameraRoundedX = math.floor(cameraX/blockSize)
    cameraRoundedY = math.floor(cameraY/blockSize)
end


function love.draw()
    local scale = 1
    local blocksX = (screenWidth/blockSize)/scale
    local blocksY = (screenHeight/blockSize)/scale

    local totalLights = 0

    love.graphics.scale(scale, scale)
    for x = 1, blocksX+2 do
        for y = 1, blocksY+2 do

            if blockContainer[x+cameraRoundedX] then
                local blockIndex = blockContainer[x+cameraRoundedX][y-cameraRoundedY]

                if blockIndex then
                    local blockRenderX = x*blockSize-(blockSize*2) - (cameraX % blockSize) -- we have a buffer for scrolling.
                    local blockRenderY = y*blockSize-(blockSize*2) + (cameraY % blockSize)
                    -- handle lighting
                    --local l = blockIndex.lighting.light -- base of 0
                    local c = {r=1,g=1,b=1}
                    local blocksBlockingDiv = 1
                    -- get above blocks
                    --[[for _, column in pairs(lightingData.sources) do
                       for _, light in pairs(column) do
                            -- calculate the distance between each other
                            local distanceX = (x + cameraRoundedX) - light.pos.x
                            local distanceY = (y - cameraRoundedY) + light.pos.y
                            local distanceSquared = math.sqrt((distanceX^2 + distanceY^2))
                    
                            local i = light.intensity
                            -- Calculate the color contribution based on the distance and intensity
                            local colorContribution = i / math.max(distanceSquared, 1) -- Use math.max instead of math.min
                    
                            -- Add the color contribution to the final color channelss

                            c.r = c.r + colorContribution * light.color.r * 4
                            c.g = c.g + colorContribution * light.color.g * 4
                            c.b = c.b + colorContribution * light.color.b * 4

                            totalLights = totalLights + 1
                        end
                    end]]
                    for i = 1, 2 do -- let's find how many blocks are above!
                        if blockContainer[x+cameraRoundedX][y-cameraRoundedY-i] then
                            blocksBlockingDiv = blocksBlockingDiv+1
                        end
                    end
                    c.r = 1--/(1+(blocksBlockingDiv^2)/1)
                    c.g = 1--/(1+(blocksBlockingDiv^2)/1)
                    c.b = 1--/(1+(blocksBlockingDiv^2)/1)

                    -- final
                    love.graphics.setColor(blockIndex.color.r*c.r, blockIndex.color.g*c.g, blockIndex.color.b*c.b)
                    --
                    love.graphics.rectangle('fill', blockRenderX, blockRenderY, blockSize, blockSize)
                    love.graphics.setColor(1,1,1)
                end
            end
        end
    end
    -- render dynamic objects / bodies
    for _, entity in pairs(entitiesModule.entities) do
        local x = entity.x - cameraX
        local y = entity.y + cameraY
        love.graphics.rectangle('fill', x-blockSize, y, entity.width, entity.height)
    end
    --[[ 
    local blockRenderX = x*blockSize-(blockSize*2) - (cameraX % blockSize) -- we have a buffer for scrolling.
    local blockRenderY = y*blockSize-(blockSize*2) + (cameraY % blockSize)
    ]]

    local cameraPosX = math.floor(cameraX/blockSize)
    local cameraPosY = math.floor(cameraY/blockSize)

    local indexX = cameraPosX+math.floor((love.mouse.getX()+(cameraX % blockSize)+blockSize)/blockSize)+1
    local indexY = (math.floor((love.mouse.getY()-(cameraY % blockSize)+blockSize)/blockSize)-cameraPosY)+1
    -- get the tile!
    love.graphics.print('Block Index X: ' .. indexX, 0, 120)
    love.graphics.print('Block Index Y: ' .. indexY, 0, 140)

    local firstTile = blocksModule.functions.getBlockAtPosition(indexX, indexY)
    --

    love.graphics.scale(1/scale, 1/scale)
    love.graphics.setColor(1,0,0)
    love.graphics.print('FPS: ' .. love.timer.getFPS())
    love.graphics.print('Seed: ' .. worldSeed, 0, 20)
    love.graphics.print('Total Lights: ' .. totalLights, 0, 40)

    love.graphics.print('Normalized Camera X: ' .. cameraRoundedX, 0, 80)
    love.graphics.print('Normalized Camera Y: ' .. cameraRoundedY, 0, 100)
    love.graphics.print('Map X: ' .. cameraX, 0, 180)
    love.graphics.print('Map Y: ' .. cameraY, 0, 200)
    love.graphics.print(firstTile and firstTile.name or 'nil', 0, 160)

    love.graphics.setColor(1,1,1)

end

love.mousepressed = function(x, y, b)
    if b == 1 then
        local cameraPosX = math.floor(cameraX/blockSize)
        local cameraPosY = math.floor(cameraY/blockSize)
        local indexX = cameraPosX+math.floor((x+(cameraX % blockSize)+blockSize)/blockSize)+1
        local indexY = (math.floor((y-(cameraY % blockSize)+blockSize)/blockSize)-cameraPosY)+1
        print('click')
        if blocksModule.world[indexX] then
            if blocksModule.world[indexX][indexY] then
                    blocksModule.world[indexX][indexY] = nil
            else
                    local block = blocksModule.functions.createBlock(5) -- get block
                    block.state = 1
                    block.slope = 0
                    block.pos = {
                        x = x,
                        y = y
                    }
                    blocksModule.world[indexX][indexY] = block
            end
        end
    end
end